# -*- coding: utf-8 -*-
"""Provide whatchlist management."""

import logging
import collections
log = logging.getLogger('msiempy')
from . import NitroDict, NitroList

class WatchlistManager(NitroList):

    def __init__(self, *args, **kwargs):
        """

        """
        super().__init__(*args, **kwargs)
        self.data=self.nitro.request('get_watchlists_no_filters', hidden=False, dynamic=False, writeOnly=False, indexedOnly=False)

        #Casting all data to Watchlist objects, better way to do it ?
        collections.UserList.__init__(self, [Watchlist(adict=item) for item in self.data if isinstance(item, (dict, NitroDict))])

    def load_details(self):
        """
        
        """
        self.perform(Watchlist.load_details, asynch=False, progress=True)


class Watchlist(NitroDict):
    """

    Complete list of watchlist fields (not values) once load with load_details()

        Dictionary keys

        name: The name of the watchlist
        type: The watchlist type
        customType: The watchlist custom type (custom field)
        dynamic: Whether this watchlist is dynamic
        hidden:  Whether this watchlist is hidden
        scored: Whether this watchlist has a scoring component (GTI for example)
        valueCount: The number of values in this watchlist
        active: Whether this watchlist is a active
        errorMsg: The error message, if there is one associated with this watchlist
        source: source
        id: The id of the watchlist
        search: A regular expression, if applicable to the type of data source
        updateType: If dynamic is true, the type of update frequency (hourly, weekly, etc)
            Accepted Values:
            EVERY_SO_MANY_MINUTES
            HOURLY_AT_SPECIFIED_MINUTE
            DAILY_AT_SPECIFIED_TIME
            WEEKLY_AT_SPECIFIED_DAYTIME
            MONTHLY_AT_SPECIFIED_DAYTIME
        updateDay: The day the watchlist should be updated, if applicable. This value will either be the day of the week (1-7, corresponding to Sunday to Saturday), or the day of the month depending on the update type.
        updateMin: If dynamic is true and a minute field is applicable to the update frequency, this will hold the minute of either the hour or day depending on updateType.
        age: The age of the watchlist values in milliseconds
        ipsid: ipsid
        recordCount: The number of records in the watchlist
        valueFile: valueFile The file that can be obtained with a call to sysGetWatchlistValues containing all of the watchlist values
        dbUrl: If an enrichment source is being set up for the watchlist, this should hold the database URL
        mountPoint: If an enrichment source is being set up for the watchlist, this would hold the mount point if applicable. See product documentation on enrichment settings for more details.
        path: If the watchlist is populated from an enrichment source, this will hold the enrichment path setting. See the enrichment configuration documentation for more details on this setting.
        port: If the watchlist is populated from an enrichment source, this will hold the enrichment port setting. See the enrichment configuration documentation for more details on this setting.
        username: If the watchlist is populated from an enrichment source, this will hold the enrichment username setting. See the enrichment configuration documentation for more details on this setting.
        password: If the watchlist is populated from an enrichment source, this will hold the enrichment password setting. See the enrichment configuration documentation for more details on this setting.
        query: If the watchlist is populated from an enrichment source, this will hold the enrichment query setting. See the enrichment configuration documentation for more details on this setting.
        lookup: If the watchlist is populated from an enrichment source, this will hold the enrichment lookup setting. See the enrichment configuration documentation for more details on this setting.
        enabled: Whether the watchlist is enabled
        jobTrackerURL: If the watchlist is populated from an enrichment source, this will hold the enrichment Job Tracker URL setting. See the enrichment configuration documentation for more details on this setting.
        jobTrackerPort: If the watchlist is populated from an enrichment source, this will hold the enrichment Job Tracker Port setting. See the enrichment configuration documentation for more details on this setting.
        postArgs: If the watchlist is populated from an enrichment source, this will hold the enrichment Post Argument setting. See the enrichment configuration documentation for more details on this setting.
        sSLCheck: If the watchlist is populated from an enrichment source, this will hold the enrichment SSL Check setting. See the enrichment configuration documentation for more details on this setting.
        ignoreRegex: If the watchlist is populated from an enrichment source, this will hold the enrichment Ignore RegEx setting. See the enrichment configuration documentation for more details on this setting.
        method: If the watchlist is populated from an enrichment source, this will hold the enrichment method setting. See the enrichment configuration documentation for more details on this setting.
        matchRegex: If the watchlist is populated from an enrichment source, this will hold the enrichment match regex setting. See the enrichment configuration documentation for more details on this setting.
        lineSkip: If the watchlist is populated from an enrichment source, this will hold the enrichment line skip setting. See the enrichment configuration documentation for more details on this setting.
        delimitRegex: If the watchlist is populated from an enrichment source, this will hold the enrichment delimit regex setting. See the enrichment configuration documentation for more details on this setting.
        groups: If the watchlist is populated from an enrichment source, this will hold the enrichment groups setting. See the enrichment configuration documentation for more details on this setting.
        values: values
    
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def add_values(self, values):
        """
        
        """
        self.nitro.request('add_watchlist_values', watchlist=self['id'], values=values)

    def data_from_id(self, id):
        """
        
        """
        info=self.nitro.request('get_watchlist_details', id=id)
        return info

    def load_details(self):
        """
        
        """
        the_id = self.data['id']
        self.data.update(self.data_from_id(the_id))
        self.data['id']=the_id

    def load_values(self, count=50):
        """
        Not properly tested yet.
        """
        self.data['values']=self.nitro.request('get_watchlist_values', id=self.data['id'], pos=0, count=count)




# -*- coding: utf-8 -*-
"""This module imports a dict into the msiem core class to provide
    a central place to aggregate API methods and parameters. The params
    are stored as docstrings to support string replacement.

    Args:
        method (str): Dict key associated with desired function
        Use normal dict access, PARAMS["method"], or PARAMS.get("method")

    Returns:
        tuple: (string, string)

        The first string is the method name that is actually used as
        the URI or passed to the ESM. The second string is the params
        required for that method. Some params require variables be
        interpolated as documented in the Attributes.

    Example:
        method, params = params["login"].format(username, password)

    Important note : Do not use sigle quotes (') to delimit data into the interpolated strings !
"""

PARAMS = {
    "login": ("login",
              """{"username": "%(username)s",
                 "password" : "%(password)s",
                 "locale": "en_US",
                 "os": "Win32"}
                 """),

    "get_devtree": ("GRP_GETVIRTUALGROUPIPSLISTDATA",
                    """{"ITEMS": "#{DC1 + DC2}",
                        "DID": "1",
                        "HD": "F",
                        "NS": "0"}
                    """),

    "get_zones_devtree": ("GRP_GETVIRTUALGROUPIPSLISTDATA",
                    """{"ITEMS": "#{DC1 + DC2}",
                        "DID": "3",
                        "HD": "F",
                        "NS": "0"}
                    """),

    "req_client_str": ("DS_GETDSCLIENTLIST",
                          """{"DSID": "%(_ds_id)s",
                              "SEARCH": ""}
                          """),

    "get_rfile": ("MISC_READFILE",
                 """{"FNAME": "%(_ftoken)s",
                 "SPOS": "0",
                 "NBYTES": "0"}
                 """),

    "get_wfile": ("MISC_WRITEFILE",
                 """{"DATA1": "%(_ds_id)s",
                 """),

                 
                 
    "map_dtree": ("map_dtree",
                  """{"dev_type": "%(dev_type)s",
                  "name": "%(ds_name)s",
                  "ds_id": "%(ds_id)s",
                  "enabled": "%(enabled)s",
                  "ds_ip": "%(ds_ip)s",
                  "hostname" : "%(hostname)s",
                  "typeID": "%(type_id)s",
                  "vendor": "",
                  "model": "",
                  "tz_id": "",
                  "date_order": "",
                  "port": "",
                  "syslog_tls": "",
                  "client_groups": "%(client_groups)s"
                  }
                  """),
                 
    "add_ds": ("dsAddDataSource", 
                """{"datasource": {
                        "parentId": {"id": "%(parent_id)s"},
                        "name": "%(name)s",
                        "id": {"id": "%(ds_id)s"},
                        "typeId": {"id": "%(type_id)s"},
                        "childEnabled": "%(child_enabled)s",
                        "childCount": "%(child_count)s",
                        "childType": "%(child_type)s",
                        "ipAddress": "%(ds_ip)s",
                        "zoneId": "%(zone_id)s",
                        "url": "%(url)s",
                        "enabled": "%(enabled)s",
                        "idmId": "%(idm_id)s",
                        "parameters": %(parameters)s
                    }}"""),

    "add_client": ("DS_ADDDSCLIENT", 
                     """{"PID": "%(parent_id)s",
                     "NAME": "%(name)s",
                     "ENABLED": "%(enabled)s",
                     "IP": "%(ds_ip)s",
                     "HOST": "%(hostname)s",
                     "TYPE": "%(type_id)s",
                     "TZID": "%(tz_id)s",
                     "DORDER": "%(dorder)s",
                     "MASKFLAG": "%(maskflag)s",
                     "PORT": "%(port)s",
                     "USETLS": "%(syslog_tls)s"
                    }"""),
                    
    "get_recs": ("devGetDeviceList?filterByRights=false",
                     """{"types": ["RECEIVER"]}
                     """),

    "get_dstypes": ("dsGetDataSourceTypes",
                     """{"receiverId": {"id": "%(rec_id)s"}
                        }
                     """),
                     
    "del_ds": ("dsDeleteDataSource",
                """{"receiverId": {"id": "%(parent_id)s"},
                    "datasourceId": {"id": "%(ds_id)s"}}
                 """),
                 
    "del_client": ("DS_DELETEDSCLIENTS",None
                    ),
                    
    "ds_last_times": ("QRY%5FGETDEVICELASTALERTTIME","""{}"""),
                      
    "zonetree": ("zoneGetZoneTree",None),
                      
    "ds_by_type": ("QRY_GETDEVICECOUNTBYTYPE",None),

   "_dev_types":  ("dev_type_map",
                    """{"1": "zone",
                        "2": "ERC",
                        "3": "datasource",
                        "4": "Database Event Monitor (DBM)",
                        "5": "DBM Database",
                        "7": "Policy Auditor",
                        "10": "Application Data Monitor (ADM)",
                        "12": "ELM",
                        "14": "Local ESM",
                        "15": "Advanced Correlation Engine (ACE)",
                        "16": "Asset datasource",
                        "17": "Score-based Correlation",
                        "19": "McAfee ePolicy Orchestrator (ePO)",
                        "20": "EPO",
                        "21": "McAfee Network Security Manager (NSM)",
                        "22": "McAfee Network Security Platform (NSP)",
                        "23": "NSP Port",
                        "24": "McAfee Vulnerability Manager (MVM)",
                        "25": "Enterprise Log Search (ELS)",
                        "254": "client_group",
                        "256": "client"}
                    """),
                    
    "ds_details": ("dsGetDataSourceDetail",
                    """{"datasourceId": 
                        {"id": "%(ds_id)s"}}
                    """),

    "get_alarms_custom_time": ("""alarmGetTriggeredAlarms?triggeredTimeRange=%(time_range)s&customStart=%(start_time)s&customEnd=%(end_time)s&status=%(status)s&pageSize=%(page_size)s&pageNumber=%(page_number)s""",
                   None),

    "get_alarms": ("""alarmGetTriggeredAlarms?triggeredTimeRange=%(time_range)s&status=%(status)s&pageSize=%(page_size)s&pageNumber=%(page_number)s""", None),

    "get_alarm_details": ("""notifyGetTriggeredNotification""", """{"id":%(id)s}"""),

    "ack_alarms": ("""alarmAcknowledgeTriggeredAlarm""", """{"triggeredIds":%(ids)s}"""),

    "unack_alarms": ("""alarmUnacknowledgeTriggeredAlarm""", """{"triggeredIds":%(ids)s}"""),

    "delete_alarms": ("""alarmDeleteTriggeredAlarm""", """{"triggeredIds":%(ids)s}"""),
    
    "get_possible_filters" : ( """qryGetFilterFields""", None ),

    "get_possible_fields" : ( """qryGetSelectFields?type=%(type)s&groupType=%(groupType)s""", None ),

    "get_esm_time" : ( """essmgtGetESSTime""",None),

    "logout" : ( """userLogout""", None ),

    "get_user_locale" : ( """getUserLocale""", None ),

    "event_query_custom_time" : ("""qryExecuteDetail?type=EVENT&reverse=false""", """{
            "config": {
                "timeRange": "%(time_range)s",
                "customStart": "%(start_time)s",
                "customEnd": "%(end_time)s",
                "fields":%(fields)s,
                "filters":%(filters)s,
                "limit":%(limit)s,
                "offset":%(offset)s
                }
                }"""),

    "event_query" : ("""qryExecuteDetail?type=EVENT&reverse=false""", """{
            "config": {
                "timeRange":"%(time_range)s",
                "fields":%(fields)s,
                "filters":%(filters)s,
                "limit":%(limit)s,
                "offset":%(offset)s
                }
                }"""),

    "query_status" : ("""qryGetStatus""", """{"resultID": %(resultID)s}"""),

    "query_result" : ("""qryGetResults?startPos=%(startPos)s&numRows=%(numRows)s&reverse=false""", """{"resultID": %(resultID)s}"""),
    
    "time_zones" : ("""userGetTimeZones""", None),

    "logout" : ("""logout""", None),
    
    "add_note_to_event" : ("""ipsAddAlertNote""", """{
        "id": {"value": "%(id)s"},
        "note": {"note": "%(note)s"}
    }"""),

    "get_watchlists_no_filters" : ("""sysGetWatchlists?hidden=%(hidden)s&dynamic=%(dynamic)s&writeOnly=%(writeOnly)s&indexedOnly=%(indexedOnly)s""", 
          None),

    "get_watchlist_details": ("""sysGetWatchlistDetails""","""{"id": %(id)s}"""),

    "add_watchlist_values": ("""sysAddWatchlistValues""","""{
        "watchlist": %(id)s,
        "values": %(values)s,
        }"""),

    "get_watchlist_values": ("""sysGetWatchlistValues?pos=%(pos)s&count=%(count)s""", """{"file": {"id": "%(id)s"}}"""),

    "get_alert_data": ("""ipsGetAlertData""", """{"id": "%(id)s"}"""),
    
    "get_sys_info"  : ("sysGetSysInfo",None),
    
    "build_stamp" : ("essmgtGetBuildStamp",None)
}


"""Provide event management.
"""

import time
import datetime
import logging
log = logging.getLogger('msiempy')

from . import NitroDict, NitroError
from .query import FilteredQueryList, FieldFilter, GroupFilter, QueryFilter
from .utils import timerange_gettimes, parse_query_result, format_fields_for_query

class EventManager(FilteredQueryList):
    """
    EventManager class.
    Interface to query and manage events.
    Inherits from FilteredQueryList.
    """ 

    #Constants
    #TODO Try grouped queries !
    TYPE='EVENT'
    GROUPTYPE='NO_GROUP'
    POSSBILE_ROW_ORDER=[
            'ASCENDING',
            'DESCENDING'
    ]

    # Declaring static value containing all the possibles
    # event fields, should be loaded once (when the session start ?)
    _possible_fields = []

    def __init__(self, fields=None, order=None, limit=None, filters=None, compute_time_range=True, *args, **kwargs):
        """
           Params
           ======
           
            fields : list of strings representing all fields you want to apprear in the Events records.
                Get the list of possible fields by calling EventManager.get_possible_fields() method.
                Some defaults fields will always be present unless removed, see (1).
            order : Not implemented yet. 
                tuple (direction, field) or a list of filters in the SIEM format.
                will set the first order according to (direction, fields).
                    -> same as using the property setter.
                If you pass a list here, will use the this raw list as the SIEM `order` field.
                Structure must be in correct format
                
                    -> same as setting _order property directly.
            limit : max number of rows per query, by default takes the value in config `default_rows` option.
            filters : list of tuple (field [values])
            compute_time_range : False if you want to send the actualy time_range in parameter for the initial query.
                Defaulted to True cause the query splitting implies computing the time_range anyway.
            *args, **kwargs : Parameters passed to `msiempy.base.FilteredQueryList.__init__()`
           
            
            Examples
            ========
            
            (1) To delete a default field
            ```
                >>>em=EventManager()
                >>>del em.fields['SrcIP']
            ```
            
            Every init parameters are also properties. E.g. :
            ```
                >>>em=EventManager(fields=['SrcIP','DstIP'],
                        order=('DESCENDING''LastTime'),
                        filters=[('DstIP','4.4.0.0/16','8.8.0.0/16')])
                >>>em.load_data()```
            Equlas :
            ```
                >>>em=EventManager()
                >>>em.fields=['SrcIP','DstIP'],
                >>>em._order=[{  "direction": 'DESCENDING',
                                "field": { "name": 'LastTime' }  }]
                >>>em.filters=[('DstIP','4.4.0.0/16','8.8.0.0/16')]
                >>>em.load_data()```
                
        """

        #Declaring attributes
        self._filters=list()
        self._order=dict()

        #Setting the default fields
        self.fields=Event.DEFAULTS_EVENT_FIELDS
        
        #Adds the specified fields and make sure there is no duplicates
        if fields :
            self.fields=list(set(self.fields+fields))

        #Set the compute_time_range propertir that is going to be used when time_range setter is called
        self.compute_time_range=compute_time_range

        #Calling super constructor : time_range set etc...
        super().__init__(*args, **kwargs)

        #Setting limit according to config or limit argument
        #TODO Try to load queries with a limit of 10k and get result as chucks of 500 with starPost nbRows
        #   and compare efficiency
        self.limit=self.requests_size if limit is None else int(limit) #IGNORE THE CONFIG 
        self.requests_size=self.limit

        if isinstance(order, list): #if a list is passed for the prder, will replace the whole param supposed in correct SIEM format
            self._order=order
        else:
            self._order=[{
                "direction": 'DESCENDING',
                "field": {
                    "name": 'LastTime'
                    }
                }]
            self.order=order #if a tuple is passed , will set the first order according to (direction, fields)

        #TODO : find a solution not to use this
        #callign super().filters=filters #https://bugs.python.org/issue14965
        super(self.__class__, self.__class__).filters.__set__(self, filters)

        #Type cast all items in the list "data" to events type objects
        self.type_cast_items_to_events()

    @property
    def table_colums(self):
        return Event.DEFAULTS_EVENT_FIELDS

    def type_cast_items_to_events(self):
        """Type cast all items in the list "data" to events type objects"""
        
        events = []
        for item in self.data:
            event = Event(item)
            events.append(event)

        self.data = events

    @property
    def order(self):
        """
        Return a list of orders representing the what the SIEM is expecting as the 'order'
        """
        return(self._order)

    @order.setter
    def order(self, order):
        """
        The order must be tuple (direction, field).
        Use _order to set with SIEM format.
        """
        if order is None :
            self._order=[{
                "direction": 'DESCENDING',
                "field": {
                    "name": 'LastTime'
                    }
                }]
        elif isinstance(order, tuple) :
            if order[0] in self.POSSBILE_ROW_ORDER:
                self._order[0]['direction']=order[0]
                self._order[0]['field']['name']=order[1]
            else:
                raise AttributeError("Illegal order value : "+str(order[0])+". The order must be in :"+str(self.POSSBILE_ROW_ORDER))
        else :
            raise AttributeError("Illegal type for argument order. Can onyl be a tuple is using the property setter. You can diretly specify a list of orders (SIEM format) by setting _order property.")

    @property
    def filters(self):
        """
        Generates the json SIEM formatted filters for the query by calling reccursive getter : config_dict.
        """
        return([f.config_dict for f in self._filters])

    def add_filter(self, afilter):
        """
        Concrete description of the FilteredQueryList method.
        It can take a tuple(fiels, [values]) or a QueryFilter sub class.
        """
        if isinstance(afilter, tuple) :
            self._filters.append(FieldFilter(afilter[0], afilter[1]))

        elif isinstance(afilter, QueryFilter) :
            self._filters.append(afilter)
        
        else :
            raise NitroError("Sorry the filters must be either a tuple(fiels, [values]) or a QueryFilter sub class.")

    def clear_filters(self):
        """
        Replace all filters by a non filtering rule.
        Acts like the is not filters.
        """ 
        self._filters=[FieldFilter('SrcIP', ['0.0.0.0/0',])]
    
    @property
    def time_range(self):
        """ Same as super class.
            Need to re-declare the time_range getter to be able to define setter.
        """
        return(super().time_range)

    @time_range.setter
    def time_range(self, time_range):
        """
        Set the time range of the query to the specified string value.
        Trys if compute_time_range is True - by default it is - to get a 
            start and a end time with utils.timerange_gettimes()
        """
        if time_range!=None and time_range!='CUSTOM' and self.compute_time_range :
            try :
                times = timerange_gettimes(time_range)
                self._time_range='CUSTOM'
                self._start_time=times[0]
                self._end_time=times[1]
            # timerange_gettimes raises AttributeError until
            # all timeranges are supported
            except AttributeError as err:
                log.warning(err)
                #TODO : find a soltuion not to use this
                #Calling super().time_range=time_range
                #https://bugs.python.org/issue14965
                super(self.__class__, self.__class__).time_range.__set__(self, time_range)
            except :
                raise
        else :
            #TODO : find a soltuion not to use this
            super(self.__class__, self.__class__).time_range.__set__(self, time_range)

    def get_possible_fields(self):
        """
        Indicate a list of possible fields that you can request in a query.
        The list is loaded from the SIEM.
        """
        return self.nitro.request('get_possible_fields', type=self.TYPE, groupType=self.GROUPTYPE)

    def _load_data(self, workers):
        """"
            Concrete helper method to execute the query and load the data : 
                -> Submit the query -> wait -> get the events -> parse -
                    -> convert to EventManager ->  set self.data and return 
            Returns a tuple ( list of Events(1) ,the status of the query )
                      tuple (items, completed).
            
            (1) aka EventManager
        """
        query_infos=dict()

        #Queries api calls are very different if the time range is custom.
        if self.time_range == 'CUSTOM' :
            query_infos=self.nitro.request(
                'event_query_custom_time',
                time_range=self.time_range,
                start_time=self.start_time,
                end_time=self.end_time,
                #order=self.order, TODO support order
                fields=format_fields_for_query(self.fields),
                filters=self.filters,
                limit=self.limit,
                offset=0, #TODO remove old ref that is useless
                includeTotal=False
                )

        else :
            query_infos=self.nitro.request(
                'event_query',
                time_range=self.time_range,
                #order=self.order, TODO support order
                fields=format_fields_for_query(self.fields),
                filters=self.filters,
                limit=self.limit,
                offset=0, #TODO remove old ref that is useless
                includeTotal=False
                )
        
        log.debug("Waiting for EsmRunningQuery object : "+str(query_infos))
        self._wait_for(query_infos['resultID'])
        events_raw=self._get_events(query_infos['resultID'])

        events=EventManager(alist=events_raw)
        #log.debug("Data loaded : "+str(events))
        
        self.data=events
        return((events,len(events)<self.limit))

    def load_data(self, **kwargs):
        """
        Specialized EventManager load_data method.
        Use super load_data implementation.
        You could decide not to use the splitting feature by 
            calling directly _load_data() 
            kwargs are passed to super().load_data()
        """
        return EventManager(alist=super().load_data(**kwargs))

    def _wait_for(self, resultID, sleep_time=0.35):
        """
        Internal method called by _load_data
        Wait and sleep - for `sleep_time` duration in seconds -
            until the query is completed
        
        TODO handle SIEM ResultUnavailable error
        """
        log.debug("Waiting for the query to be executed on the SIEM...")
        while True:
            status = self.nitro.request('query_status', resultID=resultID)
            if status['complete'] is True :
                return True
            else :
                time.sleep(sleep_time)

    def _get_events(self, resultID, startPos=0, numRows=None):
        """
        Internal method that will get the query events, 
            called by _load_data
        by default, numRows correspond to limit
        """
        
        if not numRows :
            numRows=self.limit
                
        result=self.nitro.request('query_result',
            startPos=startPos,
            numRows=numRows,
            resultID=resultID)

        #Calls a utils function to parse the [columns][rows]
        #   to format into list of dict
        #log.debug("Parsing colums : "+str(result['columns']))
        #log.debug("Parsing rows : "+str(result['rows']))
        if len(result['columns']) != len(set([column['name'] for column in result['columns']])) :
            log.error("You requested duplicated fields, the parsed fields/values results will be missmatched !")
        events=parse_query_result(result['columns'], result['rows'])
        #log.debug("Events parsed : "+str(events))
        return events
          
class Event(NitroDict):
    """
    Event.
    You can see all the requested fields and have some 
        interaction - note only - with the events
        
    Default event field keys :
        "Rule.msg",
        "Alert.SrcPort",
        "Alert.DstPort", 
        "Alert.SrcIP", 
        "Alert.DstIP", 
        "SrcMac",
        "Alert.DstMac", 
        "Alert.LastTime",
        "Rule.NormID",
        "Alert.DSIDSigID",
        "Alert.IPSIDAlertID",
        
    See msiempy/static JSON files to browse complete list of fields and filters
    """
   
    FIELDS_TABLES=["ADGroup",
        "Action",
        "Alert",
        "Asset",
        "AssetGroup",
        "AssetThreat",
        "CaseMgt",
        "CaseOrg",
        "CaseStatus",
        "Class",
        "Connection",
        "DataEnrichment",
        "GeoLoc_ASNGeoDst",
        "GeoLoc_ASNGeoSrc",
        "IOC",
        "IPS",
        "IPSCheck",
        "NDDeviceInterface_NDDevIFDst",
        "NDDeviceInterface_NDDevIFSrc",
        "NDDevice_NDDevIDDst",
        "NDDevice_NDDevIDSrc",
        "OS",
        "Rule",
        "Rule_NDSNormSigID",
        "Tag",
        "TagAsset",
        "ThirdPartyType",
        "Threat",
        "ThreatVendor",
        "TriggeredAlarm",
        "Users",
        "Vulnerability",
        "Zone_ZoneDst",
        "Zone_ZoneSrc",
        ]

   
    DEFAULTS_EVENT_FIELDS=[
        "Rule.msg",
        "Alert.SrcPort",
        "Alert.DstPort", 
        "Alert.SrcIP", 
        "Alert.DstIP", 
        "Alert.SrcMac",
        "Alert.DstMac", 
        "Alert.LastTime",
        "Rule.NormID",
        "Alert.DSIDSigID",
        "Alert.IPSIDAlertID"
        ]
    """Relatively common event fields that could be useful to have.
    """

    def __getitem__(self, key):
        """
        Automatically adding the table name of the field by iterating trought FIELDS_TABLES
        if no '.' is present in the key
        Not working. Skipping.
        TODO try with collections.UserDict.__getitem__(self, key)
        """

        return super().__getitem__(key)
        """
        if '.' not in key :
            for table in self.FIELDS_TABLES :
                try :
                    return super().__getitem__(table+'.'+key)
                except (AttributeError, KeyError) : pass
        try :
            return super().__getitem__(key)
        except (AttributeError, KeyError) : 
            if key in self.DEFAULTS_EVENT_FIELDS :
                log.error('Some default event fields are missing from SIEM reponse.')
                return 'missing'
            else:
                log.error('The SIEM dict keys are not always the same are the requested fields. check .keys()')
                raise
        """

    def clear_notes(self):
        """
        Desctructive action.
        Replace the notes by an empty string. 
        """
        NotImplementedError()

    def add_note(self, note):
        """
        Desctructive action. It's actually going to replace the note !
        Add a new note in the note field.
        """
        if len(note) >= 4000:
            log.warning("The note is longer than 4000 characters, only the first 4000 characters will be kept. The maximum accepted by the SIEM is 4096 characters.")
            note=note[:4000]+'\n\n----- MAXIMUM NOTE LENGHT REACHED, THE NOTE HAS BEEN TRUNCATED (sorry) -----'

        self.nitro.request("add_note_to_event", 
            id=self.data["Alert.IPSIDAlertID"],
            note="NOTE (msiempy-{}) : \\n{}".format(
                str(datetime.datetime.now()),
                note.replace('"','\\"').replace('\n','\\n')))
        
    def data_from_id(self, id):
        """EsmAlertData wrapper"""
        return self.nitro.request('get_alert_data', id=id)


"""Provide time ranged filtered query wrapper.
"""

import datetime
import abc
import copy
import json
import logging
log = logging.getLogger('msiempy')

from . import NitroList, NitroError, NitroObject
from .utils import format_esm_time, convert_to_time_obj, timerange_gettimes, parse_timedelta, divide_times

class FilteredQueryList(NitroList):
    """
    Base class for query based managers : AlarmManager, EventManager
    FilteredQueryList object can handle time_ranges and time splitting.
    """
    
    DEFAULT_TIME_RANGE="CURRENT_DAY"
    """
    If you don't specify any `time_range`, act like if it was "CURRENT_DAY".
    """

    POSSIBLE_TIME_RANGE=[
            "CUSTOM",
            "LAST_MINUTE",
            "LAST_10_MINUTES",
            "LAST_30_MINUTES",
            "LAST_HOUR",
            "CURRENT_DAY",
            "PREVIOUS_DAY",
            "LAST_24_HOURS",
            "LAST_2_DAYS",
            "LAST_3_DAYS",
            "CURRENT_WEEK",
            "PREVIOUS_WEEK",
            "CURRENT_MONTH",
            "PREVIOUS_MONTH",
            "CURRENT_QUARTER",
            "PREVIOUS_QUARTER",
            "CURRENT_YEAR",
            "PREVIOUS_YEAR"
    ]
    """
    List of possible time ranges : "CUSTOM",
            "LAST_MINUTE",
            "LAST_10_MINUTES",
            "LAST_30_MINUTES",
            "LAST_HOUR",
            "CURRENT_DAY",
            "PREVIOUS_DAY",
            "LAST_24_HOURS",
            "LAST_2_DAYS",
            "LAST_3_DAYS",
            "CURRENT_WEEK",
            "PREVIOUS_WEEK",
            "CURRENT_MONTH",
            "PREVIOUS_MONTH",
            "CURRENT_QUARTER",
            "PREVIOUS_QUARTER",
            "CURRENT_YEAR",
            "PREVIOUS_YEAR"
    """

    def __init__(self, time_range=None, start_time=None, end_time=None, filters=None, 
        load_async=True, requests_size=500, max_query_depth=0,
            __parent__=None, *arg, **kwargs):
        """
        Abstract base class that handles the time ranges operations, loading data from the SIEM.

            Parameters
        
            time_range : Query time range. String representation of a time range. 
                See `msiempy.base.FilteredQueryList.POSSIBLE_TIME_RANGE`
            start_time : Query starting time, can be a string or a datetime object. Parsed with dateutil.
            end_time : Query endding time, can be a string or a datetime object. Parsed with dateutil.
            filters : List of filters applied to the query.
            load_async : Load asynchonously the sub-queries. Defaulted to True.
            requests_size : number of items per request.
            max_query_depth : maximum number of supplement reccursions of division of the query times
                Meaning, if requests_size=500, slots=5 and max_query_depth=3, then the maximum capacity of 
                the list is (500*5)*(500*5)*(500*5) = 15625000000
            
        
        """

        super().__init__(*arg, **kwargs)

        #Store the query parent 
        self.__parent__=__parent__
        self.not_completed=False

        #self.nitro.config.default_rows #nb rows per request : eq limit/page_size = requests_size
        #self.nitro.config.max_rows #max nb rows 

        #Declaring attributes and types
        self._time_range=str()
        self._start_time=None
        self._end_time=None

        #self.filters=filters filter property setter should be called in the concrete class
        #TODO find a better solution to integrate the filter propertie

        self.load_async=load_async

        if start_time is not None and end_time is not None :
            self.start_time=start_time
            self.end_time=end_time
            self.time_range='CUSTOM'
        else :
            self.time_range=time_range

        self.load_async=load_async
        self.requests_size=requests_size
        self.__init_max_query_depth__=max_query_depth
        self.query_depth_ttl=max_query_depth


    @property
    def __root_parent__(self):
        """
        Internal method that return the first query of the query tree
        """
        if self.__parent__==None:
            return self
        else :
            return self.__parent__.__root_parent__

    @property
    def time_range(self):
        """
        Returns the query time range. See `msiempy.query.FilteredQueryList.POSSIBLE_TIME_RANGE`.
        Return 'CUSTOM' if internal _time_range is None and start_time annd end_time are set.
        """
        """if self.start_time is not None and self.end_time is not None :
            return('CUSTOM')
        else :"""
        return self._time_range.upper()

    @property
    def start_time(self):
        """
        Return the start time of the query in the right SIEM format.
            See `msiempy.utils.format_esm_time()`
        Use _start_time to get the datetime object
        """
        return format_esm_time(self._start_time)

    @property
    def end_time(self):
        """
        Return the end time of the query in the right SIEM format.
            See `msiempy.utils.format_esm_time()`
        Use _end_time to get the datetime object
        """
        return format_esm_time(self._end_time)

    @time_range.setter
    def time_range(self, time_range):
        """
        Set the time range of the query to the specified string value. 
        Defaulf `msiempy.queryFilteredQueryList.DEFAULT_TIME_RANGE`.
        Note : the time range is upper cased automatically.
        Throw VallueError if unrecognized time range.
        """

        if not time_range :
            self.time_range=self.DEFAULT_TIME_RANGE

        elif isinstance(time_range, str):
            time_range=time_range.upper()
            if time_range in self.POSSIBLE_TIME_RANGE :
                if time_range != 'CUSTOM':
                    self.start_time=None
                    self.end_time=None
                self._time_range=time_range
            else:
                raise ValueError("The time range must be in "+str(self.POSSIBLE_TIME_RANGE))
        else:
            raise ValueError('time_range must be a string or None')

    
    @start_time.setter
    def start_time(self, start_time):
        """
        Set the time start of the query.
        start_time can be a string or a datetime.
        If None, equivalent current_day start 00:00:00.
        """
        
        
        if isinstance(start_time, str):
            self.start_time = convert_to_time_obj(start_time)
        elif isinstance(start_time, datetime.datetime):
            self._start_time = start_time
        elif start_time==None:
             self._start_time=None#raise ValueError("Time must be string or datetime object, not None")#self.start_time = datetime.datetime.now().replace(hour=0, minute=0, second=0, microsecond=0)
        else:
            raise ValueError("Time must be string or datetime object.")
                
    
    @end_time.setter
    def end_time(self, end_time):
        """
        Set the time end of the query.
        end_time can be a string or a datetime.
        If None, equivalent now.
        """
       
        if isinstance(end_time, str):
            self.end_time = convert_to_time_obj(end_time)
        elif isinstance(end_time, datetime.datetime):
            self._end_time = end_time
        elif end_time==None:
             self._end_time=None#raise ValueError("Time must be string or datetime object, not None")
        else:
            raise ValueError("Time must be string or datetime object.")

    @abc.abstractproperty
    def filters(self):
        """ 
        Filter propertie getter. Returns a list of filters.
        TODO find a better solution to integrate the filter propertie
        """
        raise NotImplementedError()

    @filters.setter
    def filters(self, filters):
        """Query filters property : can be a list of tuple(field, [values]) 
            or just a tuple. None value will call `msiempy.query.FilteredQueryList.clear_filters()`
        Throws AttributeError if type not supported.
        """
        
        if isinstance(filters, list):
            for f in filters :
                self.add_filter(f)

        elif isinstance(filters, tuple):
            self.add_filter(filters)

        elif filters == None :
            self.clear_filters()
        
        else :
            raise AttributeError("Illegal type for the filter object, it must be a list, a tuple or None.")

    
    @abc.abstractmethod
    def add_filter(self, filter):
        """Method that figures out the way to add a filter to the query.
        """
        pass

    @abc.abstractmethod
    def clear_filters(self):
        """Method that fiures out the way to remove all filters to the query.
        """
        pass 

    @abc.abstractmethod
    def _load_data(self, workers):
        """
        Rturn a tuple (items, completed).
        completed = True if all the data that should be load is loaded.
        """
        pass

    @abc.abstractmethod
    def load_data(self, workers=15, slots=4, delta='24h'):
        """
        Method to load the data from the SIEM.
        Split the query in defferents time slots if the query apprears not to be completed.
        Splitting is done by duplicating current object, setting different times,
        and re-loading results. First your query time is split in slots of size `delta` 
        if the sub queries are not completed, divide them in the number of `slots`, this step is
        If you're looking for `max_query_depth`, it's define at the creation of the query list.

        Returns a FilteredQueryList.
        
        Note :
            IF you looking for load_async = True/False, you should pass this to the constructor method `msiempy.query.FilteredQueryList`
                or by setting the attribute manually like `manager.load_asynch=True`
            Only the first query is loaded asynchronously.

            Parameters
        
            requests_size : size (in items) for the individual requests.
            workers : numbre of parrallels task
            slots : number of time slots the query can be divided. The loading bar is 
                divided according to the number of slots
            delta : exemple : '24h', the query will be firstly divided in chuncks according to the time delta read
                with dateutil.
        
        #
        """

        items, completed = self._load_data(workers=workers)

        if not completed :
            #If not completed the query is split and items aren't actually used

            if self.query_depth_ttl > 0 :
                #log.info("The query data couldn't be loaded in one request, separating it in sub-queries...")

                if self.time_range != 'CUSTOM': #can raise a NotImplementedError if unsupported time_range
                    start, end = timerange_gettimes(self.time_range)
                else :
                    start, end = self.start_time, self.end_time

                if self.__parent__ == None and isinstance(delta, str) :
                    #if it's the first query and delta is speficied, cut the time_range in slots according to the delta
                    times=divide_times(start, end, delta=parse_timedelta(delta))
                    
                else :times=divide_times(start, end, slots=slots)
                        #IGONORING THE CONFIG ### : self.nitro.config.slots)
                
                sub_queries=list()

                #self.

                for time in times :
                    """
                    """
                    sub_query = copy.copy(self)
                    sub_query.__parent__=self
                    sub_query.compute_time_range=False
                    sub_query.time_range='CUSTOM'
                    sub_query.start_time=time[0].isoformat()
                    sub_query.end_time=time[1].isoformat()
                    sub_query.load_async=False
                    sub_query.query_depth_ttl=self.query_depth_ttl-1
                    #sub_query.requests_size=requests_size
                    sub_queries.append(sub_query)

            
                results = self.perform(FilteredQueryList.load_data, sub_queries, 
                    #The sub query is asynch only when it's set to True and it's the first query
                    asynch=False if not self.load_async else (self.__parent__==None),
                    progress=self.__parent__==None, 
                    message='Loading data from '+self.start_time+' to '+self.end_time+'. In {} slots'.format(len(times)),
                    func_args=dict(slots=slots),
                    workers=workers)

                #Flatten the list of lists in a list
                items=[item for sublist in results for item in sublist]
                
            else :
                if not self.__root_parent__.not_completed :
                    log.warning("The query won't fully complete. Try to divide in more slots or increase the requests_size")
                    self.__root_parent__.not_completed=True

        self.data=items
        return(NitroList(alist=items)) #return self ?

class QueryFilter(NitroObject):

    _possible_filters = []

    def __init__(self):
        super().__init__()

        #Setting up static constant
        """ Not checking dynamically the validity of the fields cause makes too much of unecessary requests
            self._possible_filters = self._get_possible_filters()
            """

    def get_possible_filters(self):
        """
        Return all the fields that you can filter on in a query.
        """
        return(self.nitro.request('get_possible_filters'))

    @abc.abstractproperty
    def config_dict(self):
        """
        Dump a filter in the right format.
        """
        pass

    def refresh(self):
        """
        Superclass method.
        """
        log.warning("Can't refresh filter "+str(self))

    @property
    def json(self):
        """
        Dump the filter as a json
        """
        return (json.dumps(self, indent=4, cls=NitroObject.NitroJSONEncoder))
    
    @property
    def text(self):
        """
        Text representation
        """
        return str(self.config_dict)

class GroupFilter(QueryFilter):
    """
        Based on EsmFilterGroup. See SIEM api doc.
        Used to dump groups of filters in the right format.
    """

    def __init__(self, logic='AND', *filters) :
        """
        filter : a list of filters, it can be FieldFilter or GroupFilter aka -  base.QueryFilter
        logic : 'AND' or 'OR' (i think)
        """
        super().__init__()
        
        #Declaring attributes
        self.filters=filters
        self.logic=logic

    @property
    def config_dict(self):
        """
        Could call recursively if there is other GroupFilter(s) object nested.
        Dump a filter in the right format.
        """
        return({
            "type": "EsmFilterGroup",
            "filters": [f.config_dict for f in self.filters],
            "logic":self.logic
            })
        
class FieldFilter(QueryFilter):
    """
    Based on EsmFieldFilter. See SIEM api doc.
    Used to dump a filter in the right format.
    """

    """List of possibles operators        
        """
    POSSIBLE_OPERATORS=['IN',
        'NOT_IN',
        'GREATER_THAN',
        'LESS_THAN',
        'GREATER_OR_EQUALS_THAN',
        'LESS_OR_EQUALS_THAN',
        'NUMERIC_EQUALS',
        'NUMERIC_NOT_EQUALS',
        'DOES_NOT_EQUAL',
        'EQUALS',
        'CONTAINS',
        'DOES_NOT_CONTAIN',
        'REGEX']

    """List of possible type of value and the associated keyword to pass
        to
        """
    POSSIBLE_VALUE_TYPES=[
            {'type':'EsmWatchlistValue',    'key':'watchlist'},
            {'type':'EsmVariableValue',     'key':'variable'},
            {'type':'EsmBasicValue',        'key':'value'},
            {'type':'EsmCompoundValue',     'key':'values'}]


    def __init__(self, name, values, operator='IN') :
        """
        name : field name as string
        values : list of values the field is going 
                 to be tested againts with the specified orperator
        orperator : string representing 
        """
        super().__init__()
        #Declaring attributes
        self._name=str()
        self._operator=str()
        self._values=list()
        self.name = name
        self.operator = operator
        self.values = values

    @property
    def config_dict(self):
        """
        Dump a filter in the right format.
        """
        return ({
            "type": "EsmFieldFilter",
            "field": {"name": self.name},
            "operator": self.operator,
            "values": self.values
            })

    @property
    def name(self):
        """
        Field name property getter.
        """
        return (self._name)
    
    @property
    def operator(self):
        """
        Field operator property getter.
        """
        return (self._operator)

    @property
    def values(self):
        """
        Field values property getter.
        """
        return (self._values)

    @name.setter
    def name(self, name):
        """
        Could checking dynamically the validity of the fields but turned off cause it was loading to much 
        #TODO add the list of fields check in better way and STORE the list one time only. Use class property ?
        """
        if True : # Not checking dynamically the validity of the fields cause makes too much of unecessary requests any(f.get('name', None) == name for f in self._possible_filters):
            self._name = name
        else:
            raise AttributeError("Illegal value for the "+name+" field. The filter must be in :"+str([f['name'] for f in self._possible_filters]))
       

    @operator.setter
    def operator(self, operator):
        """
        Check the value against the list of possible operators and trow error if not present.
        """
        try:
            if operator in self.POSSIBLE_OPERATORS :
                self._operator = operator
            else:
                raise AttributeError("Illegal value for the filter operator "+operator+". The operator must be in "+str(self.POSSIBLE_OPERATORS))
        except:
            raise
        
    def add_value(self, type, **args):
        """
        Add a new value to the field filter.
        
        Args could be :
            (type='EsmBasicValue',      value='a value'}. or
            (type='EsmWatchlistValue',  watchlist=1)   or 
            (type='EsmVariableValue',   variable=1}  or
            (type='EsmCompoundValue',   values=['.*']}
        """
        try:
            type_template=None
            
            #Look for the type of the object ex EsmBasicValue
            # it' used to know the type and name of value parameter we should receive next
            for possible_value_type in self.POSSIBLE_VALUE_TYPES :
                if possible_value_type['type'] == type :
                    type_template=possible_value_type
                    if type != 'EsmBasicValue' :
                        log.warning("Filtering query with other type of filter than 'EsmBasicValue' is not tested.")                            
                    break

            #Error throwing
            if type_template is not None :
                if type_template['key'] in args :
                    
                    # Adds a new value to a fields filter
                    # Filtering query with other type of filter than 'EsmBasicValue' is not tested.
                    value = args[type_template['key']]
                    if type == 'EsmBasicValue' :
                        value=str(value)
                        #log.debug('Adding a basic value to filter ('+self.text+') : '+value)
                    self._values.append({'type':type, type_template['key']:value})
                    #log.debug('The value was appended to the list: '+str(self))
                    
                #Error throwing
                else: raise KeyError ('The valid key value parameter is not present')
            else: raise KeyError ('Impossible filter')
        except KeyError as err:
            raise AttributeError("You must provide a valid named parameters containing the type and values for this filter. The type/keys must be in "+str(self.POSSIBLE_VALUE_TYPES)+"Can't be type="+str(type)+' '+str(args)+". Additionnal indicator :"+str(err) )

    def add_basic_value(self, value):
        """
        Wrapper arround add_value to add a EsmBasicValue
        """
        self.add_value(type='EsmBasicValue', value=value)

    @values.setter
    def values(self, values):
        """
        Set a list of values calls add_value if value is a 
            dict or calls add_basic_value if int, float or str
        
        """
        for val in values :
            if isinstance(val, dict):
                self.add_value(**val)

            elif isinstance(val, (int, float, str)) :
                self.add_basic_value(val)


"""Provide alarm management.
"""

import collections
import datetime
import logging
log = logging.getLogger('msiempy')

from . import NitroDict, NitroList
from .query import FilteredQueryList
from .event import EventManager, Event
from .utils import regex_match, convert_to_time_obj

class AlarmManager(FilteredQueryList):
    """
    AlarmManager class.
    Interface to query and manage Alarms.
    Inherits from FilteredQueryList.
    """
    def __init__(self, status_filter='all', page_size=None, page_number=None, filters=None, *args, **kwargs):

        """
        Params
        
            status_filter : status of the alarms to query
            page_size : max number of rows per query, by default takes the value in config `default_rows` option.
            page_number : defaulted to 1.
            filters : [(field, [values]), (field, [values])]
            fields : list of strings. Can be an EsmTriggeredAlarm or an EsmTriggeredAlarmEvent field, or any synonims. See 
            *args, **kwargs : Parameters passed to `msiempy.base.FilteredQueryList.__init__()`
            
        Examples
        
            ```
            >>>em=AlarmManager(status_filter='unacknowledged',
                filters=[('sourceIp','^10.*'), ('ruleMessage','Wordpress')]).load_data()
            ```
        """

        super().__init__(*args, **kwargs)

        #Declaring attributes
        self._alarm_filters = list(tuple())
        self._event_filters = list(tuple())
        self._status_filter = str()

        #Setting attributes
        self.status_filter=status_filter
        self.page_size=page_size if page_size is not None else self.requests_size #IGNORE THE CONFIG
        self.requests_size=self.page_size
        self.page_number=page_number if page_number is not None else 1

        #uses the parent filter setter
        #TODO : find a soltuion not to use this stinky tric
        #callign super().filters=filters #https://bugs.python.org/issue14965
        super(self.__class__, self.__class__).filters.__set__(self, filters)

        #Casting all data to Alarms objects, better way to do it ?
        collections.UserList.__init__(self, [Alarm(adict=item) for item in self.data if isinstance(item, (dict, NitroDict))])

    @property
    def table_colums(self):
        return ['id','alarmName', 'triggeredDate', 'events']

    @property
    def filters(self):
        """
        Returns the addition of alarm related filters and event related filters.
        """
        return self._alarm_filters + self._event_filters
    
    @property
    def status_filter(self):
        """
        Return the status of the alarms in the query. `status_filter` is not a filter like other cause it's computed on the SIEM side.
        Other filters are computed locally - Unlike EventManager filters.
        """
        return self._status_filter
        """
        for synonims in Alarm.POSSIBLE_ALARM_STATUS :
            if self._status_filter in synonims :
                return synonims[0]"""

    @status_filter.setter
    def status_filter(self, status_filter):
        """
        Set the status filter of the alarm query. 'acknowledged', 'unacknowledged', 'all', '' or null -> all (default is '').
        You can pass synonims of each status. See `msiempy.alarm.Alarm.POSSIBLE_ALARM_STATUS`.
        """
        status_found=False
        if type(status_filter) is str : 
            for synonims in Alarm.POSSIBLE_ALARM_STATUS :
                if status_filter in synonims:
                    self._status_filter=synonims[0]
                    status_found=True
                
        if not status_found:
            raise AttributeError("Illegal value of status filter. The status must be in "+str(Alarm.POSSIBLE_ALARM_STATUS)+' not :'+str(status_filter))

    def add_filter(self, afilter):
        """
            Make sure the filters format is tuple(field, list(values in string)).
            Takes also care of the differents synonims fields can have.
        """

        if isinstance(afilter,str):
            afilter = afilter.split('=',1)
        try:
            values = afilter[1] if isinstance(afilter[1], list) else [afilter[1]]
            values = [str(v) for v in values]
            added=False

            for synonims in Alarm.ALARM_FILTER_FIELDS :
                if afilter[0] in synonims :
                    self._alarm_filters.append((synonims[0], values))
                    added=True

            for synonims in Alarm.ALARM_EVENT_FILTER_FIELDS :
                if afilter[0] in synonims :
                    self._event_filters.append((synonims[0], values))
                    added=True

        except IndexError:
            added = False

        if not added :
            raise AttributeError("Illegal filter field value : "+afilter[0]+". The filter field must be in :"+str(Alarm.ALARM_FILTER_FIELDS + Alarm.ALARM_EVENT_FILTER_FIELDS))

    def clear_filters(self):
        """
        Reset local alarm and event filters.
        """
        self._alarm_filters = list(tuple())
        self._event_filters = list(tuple())

    def load_data(self, **kwargs):
        """
        Specialized load_data() method that convert the `msiempy.base.FilteredQueryList.load_data()` result to AlarmManager object.
        kwargs are passed to super().load_data()
        """
        return AlarmManager(alist=super().load_data(**kwargs))

    def load_events(self, workers=20, extra_fields=None, by_id=False):
        """
        Returns a new NitroList with full detailled events fields
        """
        self.perform(
            Alarm.load_events,
            asynch=True,
            workers=workers,
            progress=True,
            func_args=dict(extra_fields=extra_fields, by_id=by_id))
        return AlarmManager(alist=self)

    def _load_data(self, workers):
        """
        Concrete helper method that loads the data.
            -> Fetch the complete list of alarms -> Filter
            
        #TODO move filtering part somewhere else
        """
        alarms=list()
        if self.time_range == 'CUSTOM' :
            alarms=self.nitro.request(
                'get_alarms_custom_time',
                time_range=self.time_range,
                start_time=self.start_time,
                end_time=self.end_time,
                status=self.status_filter,
                page_size=self.page_size,
                page_number=self.page_number
                )

        else :
            alarms=self.nitro.request(
                'get_alarms',
                time_range=self.time_range,
                status=self.status_filter,
                page_size=self.page_size,
                page_number=self.page_number
                )
        #alarms = AlarmManager(alist=alarms)
        return (( self._filter(alarms, workers), len(alarms)<self.page_size ))

    def _filter(self, alarms, workers, alarmonly=False):
        """
        Helper method that filters the alarms depending on alarm and event filters.
            -> Filter dependinf on alarms related filters -> load events details
                -> Filter depending on event related filters
        Returns a AlarmsNitroList
    
        """

        alarms = AlarmManager(alist=[a for a in alarms if self._alarm_match(a)])

        if not alarmonly :
            log.info("Getting alarms infos... Please be patient.")
            detailed = self.perform(Alarm.load_details, list(alarms), asynch=True, progress=True, workers=workers)
            alarms = AlarmManager(alist=[a for a in detailed if self._event_match(a)])

        log.info(str(len(alarms)) + " alarms matching your filter(s)")
        return alarms

    def _alarm_match(self, alarm):
        """
        Filter method that is going to return True if the passed alarm match any alarm related filters.
        """
        match=True
        for alarm_filter in self._alarm_filters :
            match=False
            value = str(alarm[alarm_filter[0]]) #Can only match strings
            for filter_value in alarm_filter[1]:
                if regex_match(filter_value.lower(), value.lower()):
                    match=True
                    break
            if not match :
                break
        return match
        
    def _event_match(self, alarm):
        """
        Filter method that is going to return True if the passed alarm match any event related filters.
        """
        match=True
        for event_filter in self._event_filters :
            match=False
            values = [str(event[event_filter[0]]) for event in alarm['events']] #Can only match strings
            for filter_value in event_filter[1]:
                if any(regex_match(filter_value.lower(), value.lower()) for value in values):
                    match=True
                    break
            if not match :
                break
        return match
        
class Alarm(NitroDict):
    """
    Alarm
    Dict keys :
        id : The ID of the triggered alarm
        summary  : The summary of the triggered alarm
        assignee : The assignee for this triggered alarm
        severity : The severity for this triggered alarm
        triggeredDate : The date this alarm was triggered
        acknowledgedDate : The date this triggered alarm was acknowledged
        acknowledgedUsername : The user that acknowledged this triggered alarm
        alarmName : The name of the alarm that was triggered
        conditionType : The condition type of the alarm
        filters : The filters for this user
        queryId : The queryId for this user
        alretRateMin : The alretRateMin for this user
        alertRateCount : The alertRateCount for this user
        percentAbove : The percentAbove for this user
        percentBelow : The percentBelow for this user
        offsetMinutes : The offsetMinutes for this user
        timeFilter : The timeFilter for this user
        maximumConditionTriggerFrequency : The maximumConditionTriggerFrequency for this user
        useWatchlist : The useWatchlist for this user
        matchField : The matchField for this user
        matchValue  : The matchValue for this user
        healthMonStatus : The healthMonStatus for this user
        assigneeId : The assigneeId for this user
        escalatedDate : The escalatedDate for this user
        caseId : The caseId for this user
        caseName : The caseName for this user
        iocName : The iocName for this user
        iocId : The iocId for this user
        description : The description for this user
        actions : The actions for this user
        events : The events for this user
    
    """

    """@property
    def status(self):
        return('acknowledged' 
            if ( (len(self.data['acknowledgedDate'])>0) 
                and (len(self.data['acknowledgedUsername'])>0))
            else 'unacknowledged')"""

    POSSIBLE_ALARM_STATUS=[
        ['acknowledged', 'ack',],
        ['unacknowledged', 'unack',],
        ['', 'all', 'both']
    ]
    """
    Possible alarm statuses : 'acknowledged', 'unacknowledged' or ''
    """

    ALARM_FILTER_FIELDS = [('id',),
    ('summary','sum'),
    ('assignee','user'),
    ('severity','sever'),
    ('triggeredDate','trigdate'),
    ('acknowledgedDate','ackdate'),
    ('acknowledgedUsername','ackuser'),
    ('alarmName','name'),
    ]
    """
    Possible fields usable in a alarm filter : 'id', 'summary', 'assignee', 'severity', 'triggeredDate', 'acknowledgedDate', 'acknowledgedUsername', 'alarmName'.
    Some synonims can also be used, see source code.
    """

    ALARM_EVENT_FILTER_FIELDS=[("eventId",),
    #"severity", duplicated in ALARM_FILTER_FIELD
    #("severity", "eventseverity",),
    ("ruleMessage",'msg','rulemsg'),
    ("eventCount",'count'),
    ("sourceIp",'srcip'),
    ("destIp",'dstip'),
    ("protocol",'prot'),
    ("lastTime",'date'),
    ("eventSubType",'subtype')]
    """
    Possible fields usable in a event filter : 'eventID', 'ruleMessage', 'eventCount', 'sourceIp', 'destIp', 'protocol', 'lastTime', 'eventSubType'.
    Some synonims can also be used, see source code.

    """

    ALARM_DEFAULT_FIELDS=['triggeredDate','alarmName','status','sourceIp','destIp','ruleMessage']
    """Defaulfs fields : 'triggeredDate','alarmName','status','sourceIp','destIp','ruleMessage' (not used, may be for printing ?)
    """

    def __init__(self, *arg, **kwargs):
        """Creates a empty AlarmManager.
        """
        super().__init__(*arg, **kwargs)

    def acknowledge(self):
        """Mark the alarm as acknowledged.
        """
        self.nitro.request('ack_alarms', ids=[self.data['id']['value']])

    def unacknowledge(self):
        """Mark the alarm as unacknowledge.    
        """
        self.nitro.request('unack_alarms', ids=[self.data['id']['value']])

    def delete(self):
        """Destructive action!
        Delete the alarm.
        """
        self.nitro.request('delete_alarms', ids=[self.data['id']['value']])

    def ceate_case(self):
        """Not implemented : TODO
        """
        raise NotImplementedError()

    def load_details(self):
        """Update the alarm with detailled data loaded from the SIEM.
        """
        the_id = self.data['id']
        self.data.update(self.data_from_id(the_id))
        self.data['id']=the_id
        return self

    def refresh(self):
        """Update the alarm with detailled data loaded from the SIEM. Concrete NitroObject method.
        """
        self.load_details()

    def load_events(self, extra_fields=None, by_id=False):
        """
        Retreive the genuine Event object from an Alarm.
        extra_fields : list of extra fields. Reminder, defaul fields : `msiempy.event.Event.DEFAULTS_EVENT_FIELDS`
        by_id : will seek for event data using the getAlertData SIEM api call. (doesn't work TODO )
                If not, will use qryExecuteDetails and with a timedeal +-1second and some ip source/dest filter...
        """
        events=self.data['events']
        filters = list()
        
        if by_id==True :
            for event in events:
                try :
                    event.data=Event(id=event['Alert.IPSIDAlertID'])
                except TypeError as err:
                    #TODO Make it work !
                    log.warning("The event could not be loaded from the id : {}".format(event))
                    

        elif by_id==False:

            for event in events:
                if len(event['sourceIp']) >= 7 :
                    filters.append(('SrcIP', [str(event['sourceIp'])]))
                if len(event['destIp']) >= 7 :
                    filters.append(('DstIP', [str(event['destIp'])]))

            log.debug("Filters : "+str(filters))
            events = EventManager(
                start_time=convert_to_time_obj(events[0]['lastTime'])-datetime.timedelta(seconds=1),
                end_time=convert_to_time_obj(events[-1]['lastTime'])+datetime.timedelta(seconds=1),
                filters=filters,
                fields=extra_fields,
                max_query_depth=0,
                limit=100
            ).load_data()

            match='|'.join([event['eventId'].split('|')[1] for event in self.data['events']])
            events = EventManager(alist=events.search(match))
            self.data['events']=events

        return events

    def data_from_id(self, id):
        """

        """
        return self.nitro.request('get_alarm_details', id=str(id))

    """
    def _hasID(self):
        try :
            if self.data['id']['value'] == 0 :
                return False
            else :
                return True
        
        except KeyError :
            return False"""




# -*- coding: utf-8 -*-
"""Welcome to the msiempy base framework module documentation.
Classes listed in this module are used by other classes in sub-modules to build specialized objects and functions.
`msiempy.NitroSession` is the point of convergence of every request to the McFee ESM It provides standard dialogue with the esm.
It uses `msiempy.NitroConfig` to setup authentication, other configuration like verbosity, logfile, general timeout, are offered throught the config file.
This API offers two main types of objects to interact with the SIEM : `msiempy.NitroList`, `msiempy.NitroDict`.
`msiempy.NitroList`s have default behaviours related to parallel execution, string representation generation and search feature.
Whereas `msiempy.NitroDict` that doesn't have any specific behaviours.


Look at the class diagram : https://mfesiem.github.io/docs/msiempy/classes.png
"""
__version__ = '0.1.1'

import logging
import requests
import json
import ast
import re
import urllib.parse
import urllib3
import configparser
import os
import getpass
import abc
import collections
import tqdm
import copy
import csv
import concurrent.futures
import prettytable
from prettytable import MSWORD_FRIENDLY
import datetime
import functools
import textwrap

from .utils import regex_match, tob64
from .params import PARAMS

try :
    requests.packages.urllib3.disable_warnings(requests.packages.urllib3.exceptions.InsecureRequestWarning)
    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
except : pass

logging.getLogger("urllib3").setLevel(logging.WARNING)
log = logging.getLogger('msiempy')



class NitroError(Exception):
    """
    Base internal exception. It's used when the user/passwd is incorrect, or other specific ESM related errors.
    """
    pass

class NitroConfig(configparser.ConfigParser):
    """"Class that handles the configuration. Reads the config file whereever it is and make accessible it's values throught object properties.
    """

    CONFIG_FILE_NAME='.msiem/conf.ini'
    """.msiem/conf.ini"""
    CONF_DIR='.msiem/'
    """.msiem/"""

    CONFIG_FILE_DISCLAMER='''
        # The configuration file should be located securely in your path since it 
        # has credentials.
        # For Windows:  %APPDATA%\\\\'''+CONFIG_FILE_NAME+'''
        # For Mac :     $HOME/'''+CONFIG_FILE_NAME+'''
        # For Linux :   $XDG_CONFIG_HOME/'''+CONFIG_FILE_NAME+'''
        #        or :   $HOME/'''+CONFIG_FILE_NAME+'''
        '''
    """
        The configuration file should be located securely in your path since it 
        has credentials.
        - For Windows:  %APPDATA%\\
        - For Mac :     $HOME/
        - For Linux :   $XDG_CONFIG_HOME/ or :   $HOME/
    """

    DEFAULT_CONF_DICT={
        'esm':{'host':'', 
            'user':'',
            'passwd':''},
        'general':{'verbose':False,
            'quiet':False,
            'logfile':'',
            'timeout':30,
            'ssl_verify':False,
            'output':'text'}
    }
    """
    Default configuration file should look like this. Authentication is left empty.
    ```
    [esm]
    host = 
    user = 
    passwd = 

    [general]
    verbose = False
    quiet = False
    logfile = 
    timeout = 30
    ssl_verify = False
    output = text
    ```
    """

    def __str__(self):
        """Custom str() method that lists all config fields.
        """
        return(self.CONFIG_FILE_DISCLAMER+'\nConfiguration file : '+
            self._path+'\n'+str({section: dict(self[section]) for section in self.sections()}))

    def __init__(self, path=None, config=None, *arg, **kwarg):
        """
        Arguments
        - path: Config file speciale path, if path is left None, will automatically look for it.
        - config: Manual config dict. ex: {'general':{'verbose':True}}
        """

        super().__init__(*arg, **kwarg)

        self.read_dict(self.DEFAULT_CONF_DICT)
    
        if not path :
            self._path = self._find_ini_location()
        else : 
            self._path = path

        try :
            files=self.read(self._path)
            if len(files) == 0:
                raise FileNotFoundError

        except :
            log.info("Config file inexistant or currupted, applying defaults")

            if not os.path.exists(os.path.dirname(self._path)):
                os.makedirs(os.path.dirname(self._path))
            self.write()

        if config is not None :
            log.info("Reading config_dict : "+str(self))
            self.read_dict(config)

    def write(self):
        """
        Write the config file.
        """
        log.info("Write config file at "+self._path)
        with open(self._path, 'w') as conf:
            super().write(conf)

    def _iset(self, section, option, secure=False):
        """Internal method to interactively set  a option in a section.
        """
        msg='Enter [{}]{}'
        value = self.get(section, option)
        newvalue=''
        if option=='passwd':
            secure=True
        if secure :
            newvalue = tob64(getpass.getpass(msg.format(section, option)+'. Press <Enter> to skip: '))
        else:
            newvalue = input(msg.format(section, option)+ '. Press <Enter> to keep '+ (value if (str(value) is not '') else 'empty') + ': ')
        if newvalue != '' :
            super().set(section, option, newvalue)

    def iset(self, section, option=None, secure=False):
        """
        Will interactively set the specified section/optionby asking the user the input.
        If option stays None, all section's option will be interactively set.
        """
        if option is None :
            for key in self.options(section):
                self._iset(section, key, secure)
        else :
            self._iset(section, option, secure)

    @property
    def user(self):
        """ConfigParser.get('esm', 'user')"""
        return self.get('esm', 'user')

    @property
    def host(self):
        """ConfigParser.get('esm', 'host')"""
        return self.get('esm', 'host')

    @property
    def passwd(self):
        """ConfigParser.get('esm', 'passwd')"""
        return self.get('esm', 'passwd')

    @property
    def verbose(self):
        """ConfigParser.getboolean('general', 'verbose')"""
        return self.getboolean('general', 'verbose')

    @property
    def quiet(self):
        """ConfigParser.getboolean('general', 'quiet')"""
        return self.getboolean('general', 'quiet')

    @property
    def logfile(self):
        """ConfigParser.get('general', 'logfile')"""
        return self.get('general', 'logfile')

    @property
    def timeout(self):
        """ConfigParser.getint('general', 'timeout')"""
        return self.getint('general', 'timeout')

    @property
    def ssl_verify(self):
        """ConfigParser.getboolean('general', 'ssl_verify')"""
        return self.getboolean('general', 'ssl_verify')

    @property
    def output(self):
        """ConfigParser.get('general', 'output')"""
        return self.get('general', 'output')

   
    @staticmethod
    def _find_ini_location():
        '''
        Returns the location of a supposed conf.ini file the conf.ini file,
        If the file doesn't exist, will still return the location.
        Do not create a files not directory.
        If a .msiem/ directory exists in pwd, will return './.msiem/conf.ini'
        Or  For Windows:  %APPDATA%\\Roaming\\
            For Mac :     $HOME/
            For Linux :   $XDG_CONFIG_HOME/
                or :   $HOME/
        If your system doesn't have any of the above environment varibles,
            will return './.msiem/conf.ini'
        '''
        conf_path_dir=None

        if os.path.isdir('./'+NitroConfig.CONF_DIR):
            conf_path_dir='./'

        elif 'APPDATA' in os.environ:
                conf_path_dir = os.environ['APPDATA']

        elif 'XDG_CONFIG_HOME' in os.environ:  
            conf_path_dir = os.environ['XDG_CONFIG_HOME']

        elif 'HOME' in os.environ:  
            conf_path_dir = os.path.join(os.environ['HOME'])
            
        else:
            conf_path_dir='./'
        
        #Join configuartion filename with supposed parent directory
        conf_path=(os.path.join(conf_path_dir, NitroConfig.CONFIG_FILE_NAME))

        return(conf_path)

class NitroSession():
    '''NitroSession object represent the point of convergence of every request to the McFee ESM
    It provides standard dialogue with the esm with agument interpolation with `msiempy.params`.
    Internal `__dict__` refers to a unique instance of dict and thus, properties can be instanciated only once.
    Credentials and other configurations are read from a ./.msiem/conf.ini file.
    If the ./msiem directory desn't exists in your current directory, will assume the file is your home directory
        as ~/.msiem/conf.ini or %appdata%\\Roaming\\.msiem\\conf.ini.
    Use NitroSession.config.read() - ConfigParser object - to read a new configuration file. 
    '''

    BASE_URL = 'https://{}/rs/esm/'
    """API v2 base url.
    """

    BASE_URL_PRIV = 'https://{}/ess/'
    """Private API base URL.
    """

    __initiated__ = False
    """
    Weither the session has been intaciated. It's supposed to be a singleton.
    """
    __unique_state__ = {}
    """
    The singleton unique state.
    """
    
    config = None
    """
    NitroConfig object.
    """
    
    executor = None
    """
    Executor object.
    """
        
    def __str__(self):
        """
    """
        return repr(self.__unique_state__) 

    def __init__(self, conf_path=None, conf_dict=None):
        """
        The init method is called every time you call NitroSession() constructor.
        But the properties are actually initiated only once.
        Use logout() to trash the obejct and re instanciate NitroSession.

            Parameters

            conf_path : Configuration file path can be passed as conf_path attr.
            conf_dict : attr ie : {'esm':{'host':'myhost.com','user':'username','passwd':''}...}.
            See `msiempy.NitroConfig` class to have full details.
        """
        global log
        self.__dict__ = NitroSession.__unique_state__
        
        #Init properties only once
        if not self.__initiated__ :
            NitroSession.__initiated__ = True
            log.info('New NitroSession instance')
            
            #Private attributes
            self._headers={'Content-Type': 'application/json'}
            self._logged=False
            
            #Config parsing
            self.config = NitroConfig(path=conf_path, config=conf_dict)
            NitroSession.config=self.config

            #Set the logging configuration
            self._init_log(verbose=self.config.verbose,
                quiet=self.config.quiet,
                logfile=self.config.logfile)

    def _request(self, method, http, data=None, callback=None, raw=False, secure=False):
        """
        Helper method that format the request, handle the basic parsing of the SIEM result 
        as well as other errors.        
        If method is all upper cases, it's a private API call.
        Private API is under /ess/ and public api is under /rs/esm
        Returns None if HTTP error, Timeout or TooManyRedirects if raw=False
        """

        url=str()
        privateApiCall=False
        result=None

        #Handling private API calls formatting
        if method == method.upper():
            privateApiCall=True
            url = self.BASE_URL_PRIV
            data = self._format_params(method, **data)
            log.debug('Private API call : '+str(method)+' Formatted params : '+str(data))
        
        #Normal API calls
        else:
            url = self.BASE_URL
            if data:
                data = json.dumps(data)

        #Logging the data request if not secure | Logs anyway the method
        log.debug('Requesting HTTP '+http+' '+ method + 
            (' with data '+str(data) if (data is not None and not secure) else '') )

        try :
            result = requests.request(
                http,
                urllib.parse.urljoin(url.format(self.config.host), method),
                data=data, 
                headers=self._headers,
                verify=self.config.ssl_verify,
                timeout=self.config.timeout
            )

            if raw :
                log.debug('Returning raw requests Response object :'+str(result))
                return result

            else:
                try:
                    result.raise_for_status()

                except requests.HTTPError as e :
                    log.error(str(e)+' '+str(result.text))
                    return result.text
                    #TODO handle expired session error, result unavailable / other siem errors
                    # ERROR_InvalidFilter (228)
                    # Status Code 500: Error processing request, see server logs for more details 
                    # <Response [400]>
                    # Input Validation Error
                    # By creating a new class

                else: #
                    result = self._unpack_resp(result)

                    if privateApiCall :
                        result = self._format_priv_resp(result)

                    if callback:
                        result = callback(result)

                    log.debug('Result '+str(result)[:100]+'[...]')

                    return result

        #Soft errors
        except requests.exceptions.Timeout as e:
            log.error(e)
            return None
        except requests.exceptions.TooManyRedirects as e :
            log.error(e)
            return None
        
    def _login(self):
        """
        Internal method that will be called when the user is not logged yet.
        Throws NitroError if login fails
        """
        userb64 = tob64(self.config.user)
        passb64 = self.config.passwd
        
        resp = self.request('login', username=userb64, password=passb64, raw=True, secure=True)
        
        if resp is not None :
            if resp.status_code in [400, 401]:
                raise NitroError('Invalid username or password for the ESM')
            elif 402 <= resp.status_code <= 600:
                raise NitroError('ESM Login Error:', resp.text)
       
            self._headers['Cookie'] = resp.headers.get('Set-Cookie')
            self._headers['X-Xsrf-Token'] = resp.headers.get('Xsrf-Token')
    
            return True
        else:
            raise NitroError('ESM Login Error: Response empty')

    def request(self, request, http='post', callback=None, raw=False, secure=False, **params):
        """
            This method is the centralized interface of all request coming to the SIEM.

                Parameters

                request :   keyword corresponding to the request name in `msiempy.params` mapping.
                http :      http method.
                callback :  a callable to execute on the returned object if needed.
                raw :       if true will return the Response object from requests module.
                secure :    if true will not log the content of the request.
                **params :  interpolation parameters that will be match to `msiempy.params` templates. Check the file to be sure of the keyword arguments.

            Returns None if HTTP error, Timeout or TooManyRedirects if raw=False
        """
        log.debug("Calling nitro request : {} params={} http={} raw={} secure={} callback={}".format(
            str(request), str(params) if not secure else '***', str(http), str(raw), str(secure), str(callback)
        ))

        method, data = PARAMS.get(request)

        if data is not None :
            data =  data % params
            data = ast.literal_eval((data.replace('\n','').replace('\t','')))
           
        if method is not None:
            try :
                method = method % params
            except TypeError as err :
                if ('must be real number, not dict' in str(err)):
                    log.warning("Interpolation failed probably because of the private API calls formatting... Unexpected behaviours can happend.")

        if not self._logged and method != 'login':
            self._logged=self._login()

        try :
            return self._request(method, http, data, callback, raw, secure)

        except ConnectionError as e:
            log.critical(e)
            raise
        except Exception as e:
            log.error(e)
            raise 
        
    def logout(self):
        """ 
        This method will logout the session, clear headers and throw away the object,
            a new session will be instanciated next time.
        """
        self.request('logout', http='delete')
        self._logged=False
        NitroSession.__initiated__ = False

    @staticmethod
    def _init_log(verbose=False, quiet=False, logfile=None):
        """
        Private method. Inits the session's logger settings based on params
        All objects should be able to log stuff, so the logger is globaly accessible
        """

        log.setLevel(logging.DEBUG)

        formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')

        std = logging.StreamHandler()
        std.setLevel(logging.DEBUG)
        std.setFormatter(formatter)

        if verbose :
            std.setLevel(logging.DEBUG)
        elif quiet :
            std.setLevel(logging.CRITICAL)
        else :
            std.setLevel(logging.INFO)

        
            
            
        log.handlers=[]
        
        log.addHandler(std)

        if logfile :
            fh = logging.FileHandler(logfile)
            fh.setLevel(logging.DEBUG)
            fh.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s'))
            log.addHandler(fh)

        if verbose and quiet :
            log.warning("Verbose and quiet values are both set to True. This is a very inconsistent state. By default, verbose value has priority.")

        return (log)
    

    @staticmethod
    def _format_params(cmd, **params):
        """
        Format private API call.
        From mfe_saw project at https://github.com/andywalden/mfe_saw
        """
        params = {k: v for k, v in params.items() if v is not None}
        params = '%14'.join([k + '%13' + v + '%13' for (k, v) in params.items()])
        
        if params:
            params = 'Request=API%13' + cmd + '%13%14' + params + '%14'
        else:
            params = 'Request=API%13' + cmd + '%13%14'
        return params

    @staticmethod
    def _format_priv_resp(resp):
        """
        Format response from private API
        From mfe_saw project at https://github.com/andywalden/mfe_saw
        """
        resp = re.search('Response=(.*)', resp).group(1)
        resp = resp.replace('%14', ' ')
        pairs = resp.split()
        formatted = {}
        for pair in pairs:
            pair = pair.replace('%13', ' ')
            pair = pair.split()
            key = pair[0]
            if key == 'ITEMS':
                value = pair[-1]
            else:
                value = urllib.parse.unquote(pair[-1])
            formatted[key] = value
        return formatted

    @staticmethod
    def _unpack_resp(response) :
        """Unpack data from response.
        Args: 
            response: requests.Response response object
        Returns a list, a dict or a string
        """
        try :
            data = response.json()
            if isinstance(response.json(), dict):
                try:
                    data = data['value']
                except KeyError:
                    try:
                        data = data['return']
                    except KeyError:
                        pass
            
        except json.decoder.JSONDecodeError:
            data = response.text

        return data

class NitroObject(abc.ABC):
    """
    Base class for all nitro objects. All objects have a reference to the single `msiempy.NitroSession` object that handle the esm requests.
    """

    class NitroJSONEncoder(json.JSONEncoder):
        """
        Custom JSON encoder that will use the approprtiate propertie depending of the type of NitroObject.
        TODO return meta info about the NitroList. Maybe create a section `manager` and `data`.
        TODO support json json dumping of QueryFilers, may be by making them inherits from NitroDict.
        """
        def default(self, obj): # pylint: disable=E0202
            if isinstance(obj,(NitroDict, NitroList)):
                return obj.data
            #elif isinstance(obj, (QueryFilter)):
                #return obj.config_dict
            else:
                return json.JSONEncoder.default(self, obj) 

    @abc.abstractmethod
    def __init__(self):
        """Creates the object session.
        """
        self.nitro=NitroSession()

    def __str__(self):
        """
        str(obj) -> return text string.
        Can be a table if the object is a NitroList.
        """
        return self.text

    def __repr__(self):
        """
        repr(obj) -> return json string.
        """
        return self.json

    @abc.abstractproperty
    def text(self):
        """
        Returns printable string.
        """
        pass

    @abc.abstractproperty
    def json(self):
        """
        Returns json string representation.
        """
        pass

    @abc.abstractmethod
    def refresh(self):
        """
        Refresh the state of the object.
        """
        pass

class NitroDict(collections.UserDict, NitroObject):
    """
    Base class that represent any SIEM data that can be represented as a item of a list.
    Exemple : Event, Alarm, etc...
    Inherits from dict.
    """
    def __init__(self, adict=None, id=None):
        """
        Initiate the NitroObject and UserDict objects, load the data if id is specified, use adict agument 
        and update dict values accordingly.

            Parameters

            adict : dict object to wrap.
            id : ESM obejct unique identifier. Alert.IPSIDAlertID for exemple. 
        """
        NitroObject.__init__(self)
        collections.UserDict.__init__(self, adict)
        
        if id != None :
            self.data=self.data_from_id(id)

        if isinstance(adict, dict):
            self.data=adict

        for key in self.data :
            if isinstance(self.data[key], list):
                self.data[key]=NitroList(alist=self.data[key])

    @property
    def json(self):
        """JSON representation of a item. Basic dict.
        """
        return(json.dumps(dict(self), indent=4, cls=NitroObject.NitroJSONEncoder))

    @property
    def text(self):
        """A list of values. Not titles.
        """
        return(', '.join([str(val) for val in self.values()]))

    def refresh(self):
        """Not implemented here
        """
        log.debug('NOT Refreshing item :'+str(self)+' '+str(NotImplementedError()))

    @abc.abstractmethod
    def data_from_id(self, id):
        """This method figured out the way to retreive the item infos from an id.
        """
        pass

class NitroList(collections.UserList, NitroObject):
    """
    Base class for NitroList objects. It offers callable execution management, searcb and other data list actions.
    """

    def __init__(self, alist=None):
        """
            Parameters

            alist : list object to wrap.
        """
        NitroObject.__init__(self)
        if alist is None:
            collections.UserList.__init__(self, [])
        
        elif isinstance(alist , (list, NitroList)):
            collections.UserList.__init__(
                self, alist #[NitroDict(adict=item) for item in alist if isinstance(item, (dict, NitroDict))] 
                #Can't instanciate NitroDict, so Concrete classes has to cast the items afterwards
                )
        else :
            raise ValueError('NitroList can only be initiated based on a list')

    @property
    def table_colums(self):
        """Return the list of columns the table representation will have. This attribute is designed to overwritten.
        """            
        return []

    def _norm_dicts(self):
        """
        Internal method.
        All dict should have the same set of keys.
        Creating keys in dicts.
        """
        for item in self.data :
            if isinstance(item, (dict, NitroDict)):
                for key in self.keys :
                    if key not in item :
                        item[key]=None

    @property
    def keys(self):
        """Set of keys for all dict
        """
        #If new fields are added it won't show on text repr. Only json.
        
        manager_keys=set()
        for item in self.data:
            if isinstance(item, (dict,NitroDict)):
                manager_keys.update(item.keys())

        return manager_keys


    def get_text(self, compact=False, fields=None, max_column_width=120):
        """
        Return a acsii table string representation of the manager list

            Parameters

            compact : Returns a nice string table made with prettytable, else an '|' separated list.
            fields : list of fields you want in the table
                is None : default fields are returned by .keys attribute and sorted.
            max_column_width : when using prettytable (not compact)

        It's an expesive thing to do on big ammount of data !
        """
        
        if not fields :
            fields=sorted(self.keys)

        if len(self) == 0 :
            return('The list is empty')

        if not compact : #Table
            table = prettytable.PrettyTable()
            table.set_style(MSWORD_FRIENDLY)
            table.field_names=fields
            self._norm_dicts()

            for item in self.data:
                if isinstance(item, (dict, NitroDict)):
                    table.add_row(['\n'.join(textwrap.wrap(str(item[field]), width=max_column_width))
                        if not isinstance(item[field], NitroList)
                        else item[field].get_text() for field in fields])
                else : log.warning("Unnapropriate list element type, doesn't show on the list : {}".format(str(item)))

            if len(self.table_colums) >0 :
                try :
                    text =table.get_string(fields=self.table_colums)
                except Exception as err :
                    if "Invalid field name" in str(err):
                        text=table.get_string()
                        log.warning("Inconsistent manager state, some fields aren't present {}".format(str(err)))
                    else :
                        raise
            else: 
                text=table.get_string()

        elif compact is True :
            text='|_'
            for field in fields :
                text+=field
                text+='_|_'
            text=text[0:len(text)-1]
            text+='\n'
            for item in self.data:
                if isinstance(item, (dict, NitroDict)):
                    text+='| '
                    for field in fields :
                        if isinstance(item[field], NitroList):
                            text+=item[field].get_text(compact=True)
                        else:
                            text+=(str(item[field]))
                            text+=' | '
                    text=text[0:len(text)-1]
                else : log.warning("Unnapropriate list element type, doesn't show on the list : {}".format(str(item)))
                    #text+=textwrap.wrap(str(item),width=max_column_width)
                text+='\n'
            text=text[0:len(text)-1]
        return text



    @property
    def text(self):
        """The text properti is a shorcut to get_text() with no arguments.
        """
        return self.get_text()
        
    @property
    def json(self):
        """Dumps a JSON list of dicts representing the current list.
        """
        return(json.dumps([dict(item) for item in self.data], indent=4, cls=NitroObject.NitroJSONEncoder))

    def search(self, invert=False, match_prop='json', *pattern):
        """
        Return a list of elements that matches one or more regex patterns.
        Patterns are applied one after another. It's a logic AND.
        Use `|` inside patterns to search with logic OR.
        This method will return a new NitroList with matching data. NitroDicts in the returned NitroList do not
        references the items in the original NitroList.

            Parameters
            
            *pattern : List or string regex patterns to look for.
            invert : Weither or not to invert the search and return elements that doesn't not match search.
            match_prop : Propertie that is going to be called to search. Could be `text` or `json`.


        If you wish to apply more specific filters to NitroList list, please
        use filter(), list comprehension, or other filtering method.
            i.e. : `[item for item in list if item['cost'] > 50]`

        More on regex https://docs.python.org/3/library/re.html#re.Pattern.search
        """
        if pattern is None :
            return self
        elif len(pattern) == 0 :
            return self
        else :
            pattern=list(pattern)
            apattern=pattern.pop()
        
        matching_items=list()
        
        if isinstance(apattern, str):
            for item in self.data :
                if regex_match(apattern, getattr(item, match_prop) if isinstance(item, NitroDict) else str(item)) is not invert :
                    matching_items.append(item)
            log.debug("You're search returned {} rows : {}".format(
                len(matching_items),
                str(matching_items)[:100]+'...'))
            #Apply AND reccursively
            return NitroList(alist=matching_items).search(*pattern, invert=invert, match_prop=match_prop)
        else:
            raise ValueError('pattern must be str')

    def refresh(self):
        """
        Execute refresh function on all items.
        """
        log.warning("The function NitroList.refresh hasn't been correctly tested")
        self.perform(NitroDict.refresh)

    def perform(self, func, data=None, func_args=None, confirm=False, asynch=False,  workers=None , progress=False, message=None):
        """
        Wrapper arround executable and the data list of `msiempy.NitroList` object.
        Will execute the callable the list.

            Parameters
            
            func : callable stateless function. func is going to be called like func(item, **func_args) on all items in data.
            data : if stays None, will perform the action on all rows, else it will perfom the action on the data list.
            func_args : dict that will be passed by default to func in all calls.
            confirm : will ask interactively confirmation.
            asynch : execute the task asynchronously with `msiempy.NitroSession` executor.
            workers : mandatory if asynch is true.
            progress : to show progress bar with ETA (tqdm).
            message : To show to the user.

        Returns a list of returned results
        """

        log.debug('Calling perform func='+str(func)+
            ' data='+str(data)[:100]+
            ' func_args='+str(func_args)+
            ' confirm='+str(confirm)+
            ' asynch='+str(asynch)+
            ' workers='+str(workers)+
            ' progress='+str(progress)+
            ' message='+str(message))

        if not callable(func) :
            raise ValueError('func must be callable')

        #Confirming with user if asked
        if confirm : self._confirm_func(func, str(self))

        #Setting the arguments on the function
        func = functools.partial(func, **(func_args if func_args is not None else {}))
        
        #The data returned by function
        returned=list()

        #Usethe self contained data if not speficed otherwise
        elements=self.data
        if isinstance(data, list) and data is not None:
            elements=data
        else :
            AttributeError('data must be a list')

        #Printing message if specified.
        tqdm_args=dict()
        #The message will appear on loading bar if progress is True
        if progress is True :
            tqdm_args=dict(desc='Loading...', total=len(elements))
            if message is not None:
                tqdm_args['desc']=message

        

        #Runs the callable on list on executor or by iterating
        if asynch == True :
            if isinstance(workers, int) :
                if progress==True :
                    if not self.nitro.config.quiet:
                        #Need to call tqdm to have better support for concurrent futures executor
                        # tqdm would load the whole bar intantaneously and not wait until the callable func returns. 
                        returned=list(tqdm.tqdm(concurrent.futures.ThreadPoolExecutor(
                        max_workers=workers ).map(
                            func, elements), **tqdm_args))
                    else:
                        log.warning("You requested to show perfrom progress but config's quiet value is True, not showing tqdm load bar.")
                        returned=list(concurrent.futures.ThreadPoolExecutor(
                        max_workers=workers ).map(
                            func, elements))
                else:
                    returned=list(concurrent.futures.ThreadPoolExecutor(
                    max_workers=workers ).map(
                        func, elements))
            else:
                raise AttributeError('When asynch == True : You must specify a integer value for workers')
        else :

            if progress==True:
                if not self.nitro.config.quiet:
                    elements=tqdm.tqdm(elements, **tqdm_args)
                else:
                    log.warning("You requested to show perform progress but config's quiet value is True, not showing tqdm load bar.")

            for index_or_item in elements:
                returned.append(func(index_or_item))

        return(returned)

    @staticmethod
    def _confirm_func(func, elements):
        """
        Ask user inut to confirm the calling of `func` on `elements`.
        """
        if not 'y' in input('Are you sure you want to do this '+str(func)+' on '+
        ('\n'+str(elements) if elements is not None else 'all elements')+'? [y/n]: '):
            raise InterruptedError("The action was cancelled by the user.")